// based on the Laminar example app
//
//   https://github.com/raquo/laminar-examples/blob/master/src/main/scala/App.scala
//
// and Li Haoyi's example canvas app
//
//   http://www.lihaoyi.com/hands-on-scala-js/#MakingaCanvasApp
//

import com.raquo.laminar.api.L.{*, given}
import narr.*
import org.scalajs.dom
import org.scalajs.dom.document
import scala.math
import slash.matrix.*

class Circle(var centerX: Double, var centerY: Double, var radius: Double)

object LatticeCircleApp:
  val canvas = canvasTag(widthAttr := 600, heightAttr := 600)
  val ctx = canvas.ref.getContext("2d").asInstanceOf[dom.CanvasRenderingContext2D]
  val data = List("-1", "0", "0", "-1", "1", "0").map(Var(_))
  
  def circThru(points: Matrix[3, 2]): Option[Circle] =
    // build the matrix that maps the circle's coefficient vector to the
    // negative of the linear part of the circle's equation, evaluated at the
    // given points
    val negLinPart = Matrix.ones[3, 3]
    negLinPart.setMatrix(0, 0, points * 2.0)
    
    // find the quadrdatic part of the circle's equation, evaluated at the given
    // points
    val quadPart = Matrix[3, 1](
      NArray.tabulate[Double](3)(
        k => points(k, 0)*points(k, 0) + points(k, 1)*points(k, 1)
      )
    )
    
    // find the circle's coefficient vector, and from there its center and
    // radius
    try
      val coeffs = negLinPart.solve(quadPart)
      val centerX = coeffs(0, 0)
      val centerY = coeffs(1, 0)
      Some(Circle(
        centerX,
        centerY,
        math.sqrt(coeffs(2, 0) + centerX*centerX + centerY*centerY)
      ))
    catch
      _ => return None
  
  def draw(): Unit =
    // center and normalize the coordinate system
    val width = canvas.ref.width
    val height = canvas.ref.height
    ctx.setTransform(1.0, 0.0, 0.0, -1.0, 0.5*width, 0.5*height)
    
    // clear the previous frame
    ctx.clearRect(-0.5*width, -0.5*width, width, height)
    
    // find the resolution
    val rDisp = 5.0
    val res = width / (2.0*rDisp)
    
    // set colors
    val highlightStyle = "white"
    val gridStyle = "#404040"
    val pointFillStyles = List("#ba5d09", "#0e8a06", "#8951fb")
    val pointStrokeStyles = List("#f89142", "#58c145", "#c396fc")
    
    // draw the grid
    val rGrid = (rDisp - 0.01).floor.toInt
    val edgeScr = res * rDisp
    ctx.strokeStyle = gridStyle
    for t <- -rGrid to rGrid do
      val tScr = res * t
      
      // draw horizontal grid line
      ctx.beginPath();
      ctx.moveTo(-edgeScr, tScr)
      ctx.lineTo(edgeScr, tScr)
      ctx.stroke()
      
      // draw vertical grid line
      ctx.beginPath();
      ctx.moveTo(tScr, -edgeScr)
      ctx.lineTo(tScr, edgeScr)
      ctx.stroke()
    
    // find and draw the circle through the given points
    val dataNow = NArray.tabulate(6)(n =>
      try
        data(n).signal.now().toDouble
      catch
        _ => Double.NaN
    )
    if dataNow.forall(t => t == t.floor) then
      // all of the coordinates are integer and non-NaN
      val points = Matrix[3, 2](dataNow)
      circThru(points) match
        case Some(circ) =>
          ctx.beginPath()
          ctx.strokeStyle = highlightStyle
          ctx.arc(
            res * circ.centerX,
            res * circ.centerY,
            res * circ.radius,
            0.0, 2.0*math.Pi
          )
          ctx.stroke()
        case None =>
    
    // draw the data points
    for n <- 0 to 2 do
      val indX = 2*n
      val indY = indX + 1
      if
        dataNow(indX) == dataNow(indX).floor &&
        dataNow(indY) == dataNow(indY).floor
      then
        ctx.beginPath()
        ctx.fillStyle = pointFillStyles(n)
        ctx.strokeStyle = pointStrokeStyles(n)
        ctx.arc(
          res * dataNow(indX),
          res * dataNow(indY),
          3.0,
          0.0, 2.0*math.Pi
        )
        ctx.fill()
        ctx.stroke()
  
  def coordInput(n: Int): Input =
    input(
      typ := "number",
      cls := s"point-${(1.0 + 0.5*n).floor.toInt}",
      controlled(
        value <-- data(n).signal,
        onInput.mapToValue --> data(n).writer
      ),
      data(n).signal --> { _ => draw() }
    )
  
  def main(args: Array[String]): Unit =
    lazy val app = div(
      canvas,
      div(
        idAttr := "data-panel",
        div("x"),
        div("y"),
        coordInput(0),
        coordInput(1),
        coordInput(2),
        coordInput(3),
        coordInput(4),
        coordInput(5)
      )
    )
    renderOnDomContentLoaded(document.body, app)
